#!/usr/bin/env ruby

##
# CS50 Submit
#
# Tommy MacWilliam <tmacwilliam@cs.harvard.edu>
# David J. Malan <malan@harvard.edu>
#
#

require 'cgi'
require 'net/https'
require 'uri'
require 'optparse'

$URL_BASE = "repo.cs50.net"
$VERSION = "1.7"

# contents of .gitignore file used by submission
$GITIGNORE = %{
core
core.*
*.bmp
*.jpg
*.jpeg
*.raw
texts/
words
}

# make sure stty is re-enabled
trap("SIGINT") do
	`stty echo`
	puts ""
	exit!
end

course = ""
username = ""
project = ""
password = ""
verbose = false
very_verbose = false

OptionParser.new do |opts|
	# version flag
	opts.on( '-v', '--version', "Version" ) do
		puts $VERSION
		exit
	end

	# verbose flag
	opts.on( '-d', '--verbose', "Debugging output" ) do
		verbose = true
	end

	# very verbose flag
	opts.on( '-D', '--very-verbose', "Debugging output" ) do
		verbose = true
		very_verbose = true
	end

	# course flag
	opts.on("-c", "--course [COURSE]", "Course") do |c|
		course = c
	end

	# username flag
	opts.on("-u", "--username [USERNAME]", "Git username") do |u|
		username = u
	end

	# repository flag
	opts.on("-r", "--repo [REPOSITORY]", "Project repository") do |r|
		project = r
	end
end.parse!

# path is requried argument
if (ARGV.length != 1)
	abort "Usage: submit50 /path/to/submit"
end
path = ARGV[0]

# prompt for course
while (course == nil || course.empty?)
	print "Course: "
	course = STDIN.gets
	course = course.strip unless course == nil
end

# prompt for project
while (project == nil || project.empty?)
	print "Repository: "
	project = STDIN.gets
	project = project.strip unless project == nil
end

# prompt for username
while (username == nil || username.empty?)
	print "Username: "
	username = STDIN.gets
	username = username.strip unless username == nil
end

# prompt user for password, but don't output while typing
while (password == nil || password.empty?)
	`stty -echo`
	print "Password: "
	password = STDIN.gets
	password = password.strip unless password == nil
	`stty echo`
	puts ""
end

# lowercase and urlencode inputs
course.downcase!
project.downcase!
# remove space from usernames in particular, due to potential httpd conf syntax error 
username = CGI.escape(username.downcase.gsub(/ /, ""))
password = CGI.escape(password)

# make sure submit directory is valid
if (!File.directory?(path))
	abort "Please specify the path to the folder containing your project."
end

# make sure submit directory is not empty
if (Dir[File.join(path, '**', '*')].count { |file| File.file?(file) } == 0)
    abort "The folder you specified doesn't have any files!"
end

# ensure repo exists server-side
begin
    create_url = "https://#{$URL_BASE}/#{course}/create/#{project}/#{username}"
    uri = URI.parse(create_url)
    http = Net::HTTP.new(uri.host, uri.port)
    request = Net::HTTP::Get.new(uri.path)
    http.use_ssl = true
    http.verify_mode = OpenSSL::SSL::VERIFY_NONE
    response = http.request(request)
    response.value()
rescue
    abort "You don't seem to be connected to the Internet.\nPlease double-check your Internet connection and try again!"
end

# change to directory user specified
Dir.chdir(path) do
	# create a new git repo if one doesn't already exist
	if (!File.directory?(".git"))
		puts "Creating git repository..." if verbose
		`git init #{unless very_verbose then '--quiet > /dev/null 2>&1' end}`
	end

	# temporarily change git username to match argument
	previous_username = `git config user.name`
	`git config user.name "#{username}"`

    # temporarily move existing .gitignore file, if any, and create new one with our contents
    `mv -f .gitignore .gitignore-submit50 > /dev/null 2>&1`
    File.open('.gitignore', 'w') { |f| f.write($GITIGNORE) }

    # turn on case-insensitivity for the gitignore
    previous_ignore_case = `git config core.ignorecase`
    `git config core.ignorecase true`

	# add and commit entire directory
	puts "Committing changes..." if verbose
	`git add --all #{unless very_verbose then '> /dev/null 2>&1' end}`
	`git commit -m 'submit' --allow-empty #{unless very_verbose then '> /dev/null 2>&1' end}`

    # check return of commit, which could fail due to corrupt git directory
    commit_exitstatus = $?.exitstatus
    if (commit_exitstatus != 0)
        puts "Commit failed!" if verbose
    end

	# tag commit as submission, replacing any previous submission
	puts "Tagging changes as submission..." if verbose
	`git tag submit --force #{unless very_verbose then '> /dev/null 2>&1' end}`

    # push submission to server
	puts "Pushing changes..." if verbose
	git_url = "https://#{username}:#{password}@#{$URL_BASE}/git/#{course}/#{project}/#{username}.git"
	`git push #{git_url} --all #{unless very_verbose then '--quiet > /dev/null 2>&1' end}`

    # check return of push, would could fail for wrong credentials or non-fast-forward
    push_exitstatus = $?.exitstatus
    if (push_exitstatus != 0)
        puts "Push rejected!" if verbose
    # push succeeded, so proceed to push tags
    else
        `git push #{git_url} --tags #{unless very_verbose then '--quiet > /dev/null 2>&1' end}`
    end

    # if either push or commit failed, then we need to clone into a temporary directory and commit/push from there
	if (push_exitstatus != 0 || commit_exitstatus != 0)
		# clone remote into a temporary directory
		puts "Cloning existing repository..." if verbose

		# for now, leave the local .git as is, and repeat clone process on each submit if histories diverge
		# `rm -rf .git`

        # clone existing directory into now-empty temporary directory
		tmp_path = "/tmp/#{course}/#{project}/#{username}"
		`rm -rf #{tmp_path}`
		`mkdir -p #{tmp_path}`
		`git clone #{git_url} #{tmp_path} #{unless very_verbose then '--quiet > /dev/null 2>&1 ' end}`
	    
		# clone failed, so user must have supplied invalid credentials
		if ($?.exitstatus != 0)
			# restore previous git environment
			`git config user.name "#{previous_username}"`
			`git config core.ignorecase "#{previous_ignore_case}"`
            `mv -f .gitignore-submit50 .gitignore > /dev/null 2>&1`
			abort "Your project has NOT been submitted. Your Course, Repository\nUsername, and/or Password were incorrect. Please try again."
		end

		# clone succeeded, so replace temporary repo with new submit contents
		puts "Clone successful, re-pushing..." if verbose
		`rm -rf #{tmp_path}/*`
		`cp -r * #{tmp_path}`

		# since we cannot use local directory, try to submit from temp directory
		Dir.chdir(tmp_path) do
            # after copying files, set environment info
			`git config user.name "#{username}"`
            `mv -f .gitignore .gitignore-submit50 > /dev/null 2>&1`
            File.open('.gitignore', 'w') { |f| f.write($GITIGNORE) }

            # attempt to commit from temporary folder
			`git add --all #{unless very_verbose then '> /dev/null 2>&1' end}`
			`git commit -m 'submit' --allow-empty #{unless very_verbose then '> /dev/null 2>&1' end}`

            # check return of commit, which *should* succeed this time
            commit_exitstatus = $?.exitstatus
            if (commit_exitstatus != 0)
                puts "Commit from /tmp failed!" if verbose
            end

            # tag submission and push
			`git tag submit --force #{unless very_verbose then '> /dev/null 2>&1' end}`
			`git push #{git_url} --all #{unless very_verbose then '--quiet > /dev/null 2>&1' end}`

            # check return of push, which can fail despite successful clone if empty repo was cloned 
            push_exitstatus = $?.exitstatus
            if (push_exitstatus != 0)
                puts "Push from /tmp failed!" if verbose
            # push succeeded, so proceed to push tags
            else
                `git push #{git_url} --tags #{unless very_verbose then '--quiet > /dev/null 2>&1' end}`
            end

            # work was not actually submitted if commit or push failed
            if (commit_exitstatus != 0 || push_exitstatus != 0)
                # restore previous git environment
                `git config user.name "#{previous_username}"`
                `git config core.ignorecase "#{previous_ignore_case}"`
                `mv -f .gitignore-submit50 .gitignore > /dev/null 2>&1`
                abort "Your project has NOT been submitted. Your Course, Repository\nUsername, and/or Password were incorrect. Please try again."
            end
		end

		# for now, leave the local .git as is, and repeat clone process on each submit if histories diverge
		# previously, we moved the temp .git directory into the working directory to avoid /tmp submission again
		# `cp -r #{tmp_path}/.git .`

		# clear temp path for future submissions
		`rm -rf #{tmp_path}`
	end

    # we can only get here if some contiguous sequence of committing and pushing succeeded
	puts "Changes pushed!" if verbose
	puts "Your work HAS been submitted!"

    # restore previous git environment
    `git config user.name "#{previous_username}"`
    `git config core.ignorecase "#{previous_ignore_case}"`
    `mv -f .gitignore-submit50 .gitignore > /dev/null 2>&1`

	# check if PDF was rendered
	pdf_url = "https://#{$URL_BASE}/pdf/#{course}/#{project}/#{username}.git/submit.pdf"
	uri = URI.parse(pdf_url)

	# only show PDF url if PDF was actually rendered
	begin
		http = Net::HTTP.new(uri.host, uri.port)
		request = Net::HTTP::Get.new(uri.path)
		http.use_ssl = true
		http.verify_mode = OpenSSL::SSL::VERIFY_NONE
		request.basic_auth(username, CGI.unescape(password))
		response = http.request(request)
		response.value()
	rescue
        # pdf not found, so quit before displaying PDF url
		exit
	end

    # request to PDF succeeded, so it must exist
	puts "To view your submission, visit\n#{pdf_url}"
end
